import binascii
import io
import tarfile
from dataclasses import dataclass, field
from enum import Enum
from pathlib import Path

from fluxvault.helpers import SyncStrategy, bytes_to_human, size_of_object, tar_object
from fluxvault.log import log


@dataclass
class FluxTask:
    name: str
    params: list


@dataclass
class FileSystemEntry:
    local_path: Path
    remote_path: Path
    local_workdir: Path = field(default_factory=Path)
    local_crc: int = 0
    remote_crc: int = 0
    keeper_context: bool = True
    remote_object_exists: bool = False
    local_object_exists: bool = False
    in_sync: bool = False
    validated_remote_crc: int = 0
    file_data: bytes = b""
    sync_strategy: SyncStrategy = SyncStrategy.STRICT

    def is_dir(self) -> bool:
        return (self.local_workdir / self.local_path).is_dir()

    def expanded_remote_path(self) -> str:
        return (
            str(self.remote_path / self.local_path)
            if self.local_path != self.remote_path
            else str(self.local_path)
        )

    def crc_file(self, filename: Path, crc: int) -> int:
        with open(filename, "rb") as f:
            for chunk in iter(lambda: f.read(1024 * 128), b""):
                crc = binascii.crc32(chunk, crc)

        return crc

    def crc_directory(self, directory: Path, crc: int) -> int:
        crc = binascii.crc32(directory.name.encode(), crc)
        for path in sorted(directory.iterdir(), key=lambda p: str(p).lower()):
            crc = binascii.crc32(path.name.encode(), crc)

            if path.is_file():
                crc = self.crc_file(path, crc)
            elif path.is_dir():
                crc = self.crc_directory(path, crc)
        return crc

    def get_file_hash(self, file: Path):
        crc = self.crc_file(file, 0)
        return {str(file.relative_to(self.local_workdir)): crc}

    def get_directory_hashes(self, dir: str = "") -> dict[str, int]:
        hashes = {}

        if dir:
            p = Path(dir)
            if not p.is_absolute:
                p = self.local_workdir / self.local_path
        else:
            p = self.local_workdir / self.local_path

        if not p.exists() and p.is_dir():
            return hashes

        crc = binascii.crc32(p.name.encode())

        # this is common format "just relative path"
        hashes.update({str(p.relative_to(self.local_workdir)): crc})
        for path in sorted(p.iterdir(), key=lambda p: str(p).lower()):
            if path.is_dir():
                hashes.update(self.get_directory_hashes(str(path)))

            elif path.is_file():
                hashes.update(self.get_file_hash(path))
        return hashes

    def validate_local_object(self):
        if self.keeper_context:
            if self.local_path.is_absolute():
                raise ValueError("All paths must be relative on Keeper")

        p = self.local_workdir / self.local_path
        # print("local absolulte path", str(p))

        if not p.exists():
            p_common = self.local_workdir.parent / "common" / p.name

            if (p_common).exists():

                self.local_workdir = self.local_workdir.parent / "common"
                log.info(
                    f"Managed object {str(p)} not found locally, using file from common directory"
                )

                p = p_common
                self.local_path = p.relative_to(self.local_workdir)

            else:
                log.error(
                    f"Managed object {str(p)} not found locally or permission error... skipping!"
                )
                return

        self.local_object_exists = True
        if p.is_file():
            self.local_crc = self.crc_file(p, 0)
        elif p.is_dir():
            self.local_crc = self.crc_directory(p, 0)

    def compare_objects(self) -> dict:
        if not self.local_object_exists:
            return

        path = self.local_path
        if self.local_path != self.remote_path:
            path = self.remote_path / self.local_path
        if not self.remote_crc:  # remote file crc is 0 if it doesn't exist
            self.remote_object_exists = False
            if self.local_crc:
                log.info(f"Agent needs new object {self.local_path.name}... sending")
                return

        self.remote_object_exists = True

        if self.remote_crc != self.local_crc:
            self.in_sync = False
            if (
                self.validated_remote_crc == self.remote_crc
                and self.local_object_exists
            ):
                log.info(
                    f"Agent remote object {str(path)} is different than local object but has been validated due to sync strategy"
                )
                return

            if self.local_object_exists:
                log.info(
                    f"Agent remote object {str(path)} is different that local object... analyzing further"
                )
                return

        if self.remote_crc == self.local_crc:
            log.info(f"Agent object {str(path)} is up to date... skipping!")
            self.in_sync = True


@dataclass
class FileSystemeGroup:
    managed_objects: list[FileSystemEntry] = field(default_factory=list)
    root_dir: Path = field(default_factory=Path)
    flattened: bool = False

    def __iter__(self):
        yield from self.managed_objects

    @classmethod
    def filter_hierarchy(
        cls, current_path: Path, existing_paths: list[Path]
    ) -> list[Path]:
        for existing_path in existing_paths.copy():
            if current_path.is_relative_to(existing_path):
                # our ancestor is already in the list. We will get replaced
                # when they get synced (don't add ourselves)
                continue

            elif existing_path.is_relative_to(current_path):
                # we are higher up the tree.. remove existing_path and
                # install ourselves
                existing_paths.remove(existing_path)
                existing_paths.append(current_path)

        return existing_paths

    def expanded_remote_paths(self) -> list[str]:
        return [
            str(x.remote_path / x.local_path)
            if x.local_path != x.remote_path
            else str(x.local_path)
            for x in self.managed_objects
        ]

    def merge_config(self, objects):
        for obj in objects:
            if isinstance(obj, FileSystemEntry):
                self.managed_objects.append(obj)
            else:
                log.error(
                    f"Object of type {type(obj)} added to file manager, must be `FileSystemEntry`"
                )

    def add_objects(self, objects: list):
        for obj in objects:
            self.add_object(obj)

    def add_object(self, obj: FileSystemEntry):
        obj.local_workdir = self.root_dir
        self.managed_objects.append(obj)

    def get_object_by_remote_path(self, remote: Path) -> FileSystemEntry:
        for obj in self.managed_objects:
            path = (
                obj.remote_path / obj.local_path
                if obj.local_path != obj.remote_path
                else obj.local_path
            )

            if path == remote:
                return obj

    def get_all_objects(self) -> list[FileSystemEntry]:
        return self.managed_objects

    # ToDo: Rename
    def objects_to_agent_list(self) -> list:
        # ToDo: Composition
        objects = []
        for obj in self.managed_objects:
            if obj.local_object_exists and (
                not obj.remote_object_exists or not obj.in_sync
            ):
                if (
                    obj.remote_object_exists
                    and obj.sync_strategy == SyncStrategy.ENSURE_CREATED
                ):
                    continue

                path_str = (
                    str(obj.remote_path / obj.local_path)
                    if obj.local_path != obj.remote_path
                    else str(obj.local_path)
                )
                uncompressed = False
                is_dir = False

                abs_local_path = obj.local_workdir / obj.local_path
                if not obj.is_dir():
                    ONE_MB = 1048576 * 1

                    size = size_of_object(abs_local_path)
                    if size > ONE_MB:
                        log.info(
                            f"File {obj.local_path} with size {bytes_to_human(size)} is larger than uncompressed limit... compressing"
                        )
                        data = tar_object(abs_local_path)
                    else:  # < 1MB
                        uncompressed = True
                        with open(abs_local_path, "rb") as f:
                            data = f.read()

                if obj.is_dir():
                    is_dir = True
                    is_empty = not any((abs_local_path).iterdir())
                    if is_empty:
                        data = ""
                    else:  # we need to tar up son
                        fh = io.BytesIO()
                        # lol, hope the files aren't too big, or, you got plenty of ram
                        with tarfile.open(fileobj=fh, mode="w|bz2") as tar:
                            tar.add(
                                self.root_dir / obj.local_path,
                                # arcname=obj.local_path.name,
                                arcname="",
                            )
                        data = fh.getvalue()
                objects.append(
                    {
                        "path": path_str,
                        "data": data,
                        "is_dir": is_dir,
                        "uncompressed": uncompressed,
                    }
                )
        return objects

    def update_root_dir(self, dir: Path):
        self.root_dir = dir
        for obj in self.managed_objects:
            obj.local_workdir = self.root_dir
            obj.validate_local_object()


@dataclass
class FluxComponentConfig:
    name: str
    file_manager: FileSystemeGroup = field(default_factory=FileSystemeGroup)
    tasks: list[FluxTask] = field(default_factory=list)
    root_dir: Path = field(default_factory=Path)
    working_dir: Path = Path("/tmp")

    def update_root_dir(self, dir: Path):
        # change name to update dirs
        # root dir referes to where the physical files are
        # working dir is where the remote end puts them
        self.root_dir = dir
        self.file_manager.update_root_dir(self.root_dir)

    def add_tasks(self, tasks: list):
        for task in tasks:
            self.add_task(task)

    def add_task(self, task: FluxTask):
        self.tasks.append(task)

    def get_task(self, name) -> FluxTask:
        for task in self.tasks:
            if task.name == name:
                return task


@dataclass
class FluxAppConfig:
    name: str
    components: list[FluxComponentConfig] = field(default_factory=list)
    comms_port: int = 8888
    sign_connections: bool = False
    signing_key: str = ""
    polling_interval: int = 900
    run_once: bool = False
    root_dir: Path = field(default_factory=Path)
    agent_ips: list[str] = field(default_factory=list)
    file_manager: FileSystemeGroup = field(default_factory=FileSystemeGroup)

    def add_component(self, component: FluxComponentConfig):
        existing = next(
            filter(lambda x: x.name == component.name, self.components), None
        )
        if existing:
            log.warn(f"Component already exists: {component.name}")
            return

        component.root_dir = self.root_dir / component.name
        self.merge_global_into_component(component)
        self.components.append(component)

    def ensure_included(self, name: str) -> FluxComponentConfig:
        component = next(filter(lambda x: x.name == name, self.components), None)
        if not component:
            component = FluxComponentConfig(name)
            self.add_component(component)

        return component

    def get_component(self, name: str) -> FluxComponentConfig:
        return next(filter(lambda x: x.name == name, self.components), None)

    def merge_global_into_component(self, component: FluxComponentConfig):
        global_config = self.file_manager.get_all_objects()
        component.file_manager.merge_config(global_config)

    def ensure_removed(self, name: str):
        self.components = [c for c in self.components if c.get("name") != name]

    def update_common_objects(self, files: list[FileSystemEntry]):
        self.file_manager.add_objects(files)

    def update_root_dir(self, dir: Path):
        self.root_dir = dir
        for component in self.components:
            component.update_root_dir(self.root_dir / component.name)
        self.file_manager.update_root_dir(self.root_dir / "common")



import asyncio
import getpass
import logging
from pathlib import Path

import keyring
import typer
import yaml

from fluxvault import FluxAgent, FluxKeeper
from fluxvault.fluxapp_config import (
    FileSystemEntry,
    FluxAppConfig,
    FluxComponentConfig,
    FluxTask,
)
from fluxvault.helpers import SyncStrategy
from fluxvault.registrar import FluxAgentRegistrar, FluxPrimaryAgent

PREFIX = "FLUXVAULT"

app = typer.Typer(rich_markup_mode="rich", add_completion=False)
log = logging.getLogger("fluxvault")


class colours:
    HEADER = "\033[95m"
    OKBLUE = "\033[94m"
    OKCYAN = "\033[96m"
    OKGREEN = "\033[92m"
    WARNING = "\033[93m"
    FAIL = "\033[91m"
    ENDC = "\033[0m"
    BOLD = "\033[1m"
    UNDERLINE = "\033[4m"


def configure_logs(log_to_file, logfile_path, debug):
    vault_log = logging.getLogger("fluxvault")
    fluxrpc_log = logging.getLogger("fluxrpc")
    level = logging.DEBUG if debug else logging.INFO

    formatter = logging.Formatter(
        "%(asctime)s: fluxvault: %(levelname)s: %(message)s", "%Y-%m-%d %H:%M:%S"
    )

    vault_log.setLevel(level)
    fluxrpc_log.setLevel(level)

    stream_handler = logging.StreamHandler()
    stream_handler.setFormatter(formatter)
    file_handler = logging.FileHandler(logfile_path, mode="a")
    file_handler.setFormatter(formatter)

    vault_log.addHandler(stream_handler)
    fluxrpc_log.addHandler(stream_handler)
    if log_to_file:
        fluxrpc_log.addHandler(file_handler)
        vault_log.addHandler(file_handler)


def yes_or_no(question, default="yes"):
    """Ask a yes/no question via input() and return their answer.

    "question" is a string that is presented to the user.
    "default" is the presumed answer if the user just hits <Enter>.
            It must be "yes" (the default), "no" or None (meaning
            an answer is required of the user).

    The "answer" return value is True for "yes" or False for "no".
    """
    valid = {"yes": True, "y": True, "ye": True, "no": False, "n": False}
    if default is None:
        prompt = " [yes/no] "
    elif default == "yes":
        prompt = f" [{colours.OKGREEN}Yes{colours.ENDC}] "
    elif default == "no":
        prompt = f" [{colours.OKGREEN}No{colours.ENDC}] "
    else:
        raise ValueError("invalid default answer: '%s'" % default)

    while True:
        print(question + prompt, end="")
        choice = input().lower()
        if default is not None and choice == "":
            return valid[default]
        elif choice in valid:
            return valid[choice]
        else:
            print("Please respond with 'yes' or 'no' " "(or 'y' or 'n').\n")


def get_signing_key(signing_address) -> str:
    signing_key = keyring.get_password("fluxvault_app", signing_address)

    if not signing_key:
        signing_key = getpass.getpass(
            f"\n{colours.OKGREEN}** WARNING **\n\nYou are about to enter your private key into a 3rd party application. Please make sure your are comfortable doing so. If you would like to review the code to make sure your key is safe... please visit https://github.com/RunOnFlux/FluxVault to validate the code.{colours.ENDC}\n\n Please enter your private key (in WIF format):\n"
        )
        store_key = yes_or_no(
            "Would you like to store your private key in your device's secure store?\n\n(macOS: keyring, Windows: Windows Credential Locker, Ubuntu: GNOME keyring.\n\n This means you won't need to enter your private key every time this program is run.",
        )
        if store_key:
            keyring.set_password("fluxvault_app", signing_address, signing_key)

    return signing_key


def build_app_from_cli(
    app_name,
    managed_objects,
    sign_connections,
    signing_address,
    agent_ips,
    run_once,
    polling_interval,
    comms_port,
) -> FluxAppConfig:
    if sign_connections:
        signing_address = get_signing_key()
        if not signing_address:
            raise ValueError(
                "signing_address must be provided if signing connections (keyring)"
            )

    app = FluxAppConfig(
        app_name,
        sign_connections=sign_connections,
        signing_key=signing_address,
        agent_ips=agent_ips,
        run_once=run_once,
        polling_interval=polling_interval,
        comms_port=comms_port,
    )

    common_objects = []
    for obj_str in managed_objects:
        parts = obj_str.split("@")

        component_name = ""
        if len(parts) > 1:
            component_name = parts[1]
            obj_str = parts[0]

        split_obj = obj_str.split(":")
        print("split", split_obj)
        local = Path(split_obj[0])

        sync_strat = None
        try:
            remote = Path(split_obj[1])
            print("remote", remote)
            # this will break on remote paths of S, A, or C"
            if str(remote) in ["S", "A", "C"]:
                # we don't have a remote, just a sync strat
                sync_strat = remote
                remote = local
        except IndexError:
            # we don't have a remote path
            remote = local
        print("strat", sync_strat)
        print("remote", remote)
        if not sync_strat:
            try:
                sync_strat = Path(split_obj[2])
            except IndexError:
                sync_strat = "S"

        match sync_strat:
            case "S":
                sync_strat = SyncStrategy.STRICT
            case "A":
                sync_strat = SyncStrategy.ALLOW_ADDS
            case "C":
                sync_strat = SyncStrategy.ENSURE_CREATED

        if local.name != str(local):
            log.error(f"Local file absolute path not allowed for: {local}... skipping")
            continue

        managed_object = FileSystemEntry(local, remote, sync_strategy=sync_strat)

        if not component_name:
            common_objects.append(managed_object)
            continue

        component = app.ensure_included(component_name)
        component.file_manager.add_object(managed_object)

    app.update_common_objects(common_objects)

    return app


def managed_objects_builder(files: list) -> list:
    managed_objects = []
    for file in files:
        local = Path(file.get("name", "."))
        remote = Path(file.get("remote_path", "."))
        sync_strategy = SyncStrategy[
            file.get("sync_strategy", SyncStrategy.STRICT.name)
        ]

        managed_object = FileSystemEntry(local, remote, sync_strategy=sync_strategy)
        managed_objects.append(managed_object)
    return managed_objects


# do this as a lambda?
# flux_tasks = []
# map(lambda x: flux_tasks.append(FluxTask(x.get("name"), x.get("params"))), tasks)
def tasks_builder(tasks: list) -> list:
    flux_tasks = []
    for task in tasks:
        flux_task = FluxTask(task.get("name"), task.get("params"))
        flux_tasks.append(flux_task)
    return flux_tasks


def build_apps_from_loadout_file(path: str) -> list:

    loadout_path = Path(path)
    apps = []

    if not loadout_path.exists():
        raise ValueError(f"Loadout path {loadout_path} doesn't exist")

    with open(str(loadout_path), "r") as stream:
        try:
            loadout = yaml.safe_load(stream)
        except yaml.YAMLError as exc:
            raise ValueError(f"loadout {loadout_path} is not parseable.\n\n{exc}")

    for app_name, params in loadout.get("apps").items():
        components = params.pop("components", [])
        # agent_ips = params.get("agent_ips")
        # polling_interval = params.get("polling_interval")
        common_objects = params.pop("managed_objects", [])
        app = FluxAppConfig(app_name, **params)
        app.file_manager.add_objects(managed_objects_builder(common_objects))

        for component_name, directives in components.items():
            component = FluxComponentConfig(component_name)
            for directive, items in directives.items():
                match directive:
                    case "working_dir":
                        component.working_dir = items
                    case "managed_objects":
                        component.file_manager.add_objects(
                            managed_objects_builder(items)
                        )
                    case "tasks":
                        component.add_tasks(tasks_builder(items))
            app.add_component(component)
        apps.append(app)
    return apps


@app.command()
def keeper(
    comms_port: int = typer.Option(
        8888,
        "--comms-port",
        "-p",
        envvar=f"{PREFIX}_COMMS_PORT",
        show_envvar=False,
    ),
    app_name: str = typer.Option(
        None,
        "--app-name",
        "-a",
        envvar=f"{PREFIX}_APP_NAME",
        show_envvar=False,
    ),
    vault_dir: str = typer.Option(
        None,
        "--vault-dir",
        "-d",
        envvar=f"{PREFIX}_VAULT_DIR",
        show_envvar=False,
    ),
    loadout_path: str = typer.Option(
        None,
        "--loadout-path",
        "-l",
        envvar=f"{PREFIX}_LOADOUT_PATH",
        show_envvar=False,
    ),
    managed_objects: str = typer.Option(
        "",
        "--managed-objects",
        "-m",
        envvar=f"{PREFIX}_MANAGED_OBJECTS",
        show_envvar=False,
        help="""Comma seperated string of managed object paths.
        
        Local files must be a relative path (relative to vault_dir)
        Remote files can be relative (working_dir) or absolute
        
        If using local / remote files, file name must match

        Any remote directories will be created if they don't exist

        Example:

        --managed-files file1.py,file2.txt:/remote/path/file2.txt,file3.py:dir/file3.py
        """,
    ),
    polling_interval: int = typer.Option(
        300,
        "--polling-interval",
        "-i",
        envvar=f"{PREFIX}_POLLING_INTERVAL",
        show_envvar=False,
    ),
    run_once: bool = typer.Option(
        False,
        "--run-once",
        "-o",
        envvar=f"{PREFIX}_RUN_ONCE",
        show_envvar=False,
        help="Contact agents once and exit",
    ),
    agent_ips: str = typer.Option(
        "",
        envvar=f"{PREFIX}_AGENT_IPS",
        show_envvar=False,
        help="If your not using app name to determine ips",
    ),
    sign_connections: bool = typer.Option(
        False,
        "--sign-connections",
        "-q",
        envvar=f"{PREFIX}_SIGN_CONNECTIONS",
        show_envvar=False,
        help="Whether or not to sign outbound connections",
    ),
    signing_address: str = typer.Option(
        "",
        envvar=f"{PREFIX}_SIGNING_ADDRESS",
        show_envvar=False,
        help="This is used to associate private key in keychain",
    ),
    gui: bool = typer.Option(
        False,
        "--gui",
        "-g",
        envvar=f"{PREFIX}_GUI",
        show_envvar=False,
        hidden=True,
        help="Run local gui server",
    ),
):

    if not vault_dir:
        vault_dir = Path().home() / ".vault"

    agent_ips = agent_ips.split(",")
    agent_ips = list(filter(None, agent_ips))

    managed_objects = managed_objects.split(",")
    managed_objects = list(filter(None, managed_objects))

    apps_config = []

    # this ignores any other command line directive
    if loadout_path:
        configs = build_apps_from_loadout_file(loadout_path)
        apps_config.extend(configs)

    # configure single app via command line parameters, Must have managed_objects
    elif app_name:
        if not managed_objects:
            raise ValueError(
                "If running single app (if app_name is set) you must pass in managed-files via cli param, alternatively, try using a loadout"
            )
        config = build_app_from_cli(
            app_name,
            managed_objects,
            sign_connections,
            signing_address,
            agent_ips,
            run_once,
            polling_interval,
            comms_port,
        )
        apps_config.append(config)

    else:
        raise ValueError("Either app_name or loadout_path must be specified")
    # Make sure all params are getting passed into app config first (ports etc)

    flux_keeper = FluxKeeper(
        apps_config=apps_config,
        vault_dir=vault_dir,
        gui=gui,
    )

    async def main():
        await flux_keeper.manage_apps()

    loop = asyncio.get_event_loop()
    loop.run_until_complete(main())


@app.command()
def agent(
    bind_address: str = typer.Option(
        "0.0.0.0",
        "--bind-address",
        "-b",
        envvar=f"{PREFIX}_BIND_ADDRESS",
        show_envvar=False,
    ),
    bind_port: int = typer.Option(
        8888,
        "--bind-port",
        "-p",
        envvar=f"{PREFIX}_BIND_PORT",
        show_envvar=False,
    ),
    enable_registrar: bool = typer.Option(
        False,
        "--registrar",
        "-s",
        envvar=f"{PREFIX}_REGISTRAR",
        show_envvar=False,
        help="Act as a proxy registrar for other agents",
    ),
    registrar_port: int = typer.Option(
        "2080",
        "--registrar-port",
        "-z",
        envvar=f"{PREFIX}_REGISTRAR_PORT",
        show_envvar=False,
        help="Port for registrar to listen on",
    ),
    registrar_address: str = typer.Option(
        "0.0.0.0",
        "--registrar-address",
        "-v",
        envvar=f"{PREFIX}_REGISTRAR_ADDRESS",
        show_envvar=False,
        help="Address for registrar to bind on",
    ),
    enable_registrar_fileserver: bool = typer.Option(
        False,
        "--registrar-fileserver",
        "-q",
        envvar=f"{PREFIX}_REGISTRAR_FILESERVER",
        show_envvar=False,
        help="Serve files over http (no authentication)",
    ),
    working_dir: str = typer.Option(
        "/tmp",
        "--working-dir",
        "-i",
        envvar=f"{PREFIX}_WORKING_DIR",
        show_envvar=False,
        help="Where files will be stored",
    ),
    whitelisted_addresses: str = typer.Option(
        "",
        "--whitelist-addresses",
        "-w",
        envvar=f"{PREFIX}_WHITELISTED_ADDRESSES",
        show_envvar=False,
        help="Comma seperated addresses to whitelist",
    ),
    verify_source_address: bool = typer.Option(
        False,
        "--verify-source-address",
        "-z",
        envvar=f"{PREFIX}_VERIFY_SOURCE_ADDRESS",
        show_envvar=False,
        help="Matches source ip to your whitelist",
    ),
    signed_vault_connections: bool = typer.Option(
        False,
        "--signed-vault-connections",
        "-k",
        envvar=f"{PREFIX}_SIGNED_VAULT_CONNECTIONS",
        show_envvar=False,
        help="Expects all keeper connections to be signed",
    ),
    zelid: str = typer.Option(
        "",
        envvar=f"{PREFIX}_ZELID",
        show_envvar=False,
        help="Testing only... if you aren't running this on a Fluxnode",
    ),
    subordinate: bool = typer.Option(
        False,
        "--subordinate",
        envvar=f"{PREFIX}_SUBORDINATE",
        show_envvar=False,
        help="If this agent is a subordinate of another agent",
    ),
    primary_agent_name: str = typer.Option(
        "fluxagent",
        "--primary-agent-name",
        envvar=f"{PREFIX}_PRIMARY_AGENT_NAME",
        show_envvar=False,
        help="Primary agent name",
    ),
    primary_agent_address: str = typer.Option(
        "",
        "--primary-agent-address",
        envvar=f"{PREFIX}_PRIMARY_AGENT_ADDRESS",
        show_envvar=False,
        hidden=True,
        help="Primary agent address",
    ),
    primary_agent_port: int = typer.Option(
        "2080",
        "--primary-agent-port",
        envvar=f"{PREFIX}_PRIMARY_AGENT_PORT",
        show_envvar=False,
        hidden=True,
        help="Primary agent port",
    ),
):

    whitelisted_addresses = whitelisted_addresses.split(",")
    whitelisted_addresses = list(filter(None, whitelisted_addresses))

    registrar = None
    if enable_registrar:
        registrar = FluxAgentRegistrar(
            bind_address=registrar_address,
            bind_port=registrar_port,
            enable_fileserver=enable_registrar_fileserver,
        )

    primary_agent = None
    if subordinate:
        primary_agent = FluxPrimaryAgent(
            name=primary_agent_name,
            address=primary_agent_address,
            port=primary_agent_port,
        )

    agent = FluxAgent(
        bind_address=bind_address,
        bind_port=bind_port,
        enable_registrar=enable_registrar,
        registrar=registrar,
        primary_agent=primary_agent,
        working_dir=working_dir,
        whitelisted_addresses=whitelisted_addresses,
        verify_source_address=verify_source_address,
        signed_vault_connections=signed_vault_connections,
        zelid=zelid,
        subordinate=subordinate,
    )

    agent.run()


@app.callback()
def main(
    debug: bool = typer.Option(
        False,
        "--debug",
        envvar=f"{PREFIX}_DEBUG",
        show_envvar=False,
        help="Enable extra debug logging",
    ),
    enable_logfile: bool = typer.Option(
        False,
        "--log-to-file",
        "-l",
        envvar=f"{PREFIX}_ENABLE_LOGFILE",
        show_envvar=False,
        help="Turn on logging to file",
    ),
    logfile_path: str = typer.Option(
        "/tmp/fluxvault.log",
        "--logfile-path",
        "-p",
        envvar=f"{PREFIX}_LOGFILE_PATH",
        show_envvar=False,
    ),
):
    configure_logs(enable_logfile, logfile_path, debug)


@app.command()
def remove_private_key(zelid: str):
    try:
        keyring.delete_password("fluxvault_app", zelid)
    except keyring.errors.PasswordDeleteError:
        typer.echo("Private key doesn't exist")
    else:
        typer.echo("Private key deleted")


def entrypoint():
    """Called by console script"""
    app()


if __name__ == "__main__":
    app()



from __future__ import annotations

import asyncio
import binascii
import importlib
import io
import os
import pty
import ssl
import subprocess
import sys
import tarfile
import tempfile
from pathlib import Path

import aiofiles
import aioshutil
from aiofiles import os as aiofiles_os
from aiohttp import ClientSession
from fluxrpc.auth import SignatureAuthProvider
from fluxrpc.protocols.jsonrpc import JSONRPCProtocol
from fluxrpc.server import RPCServer
from fluxrpc.transports.socket.server import EncryptedSocketServerTransport
from cryptography import x509
from cryptography.hazmat.primitives import hashes
from cryptography.hazmat.primitives.asymmetric import rsa
from cryptography.hazmat.primitives.serialization import (
    Encoding,
    NoEncryption,
    PrivateFormat,
)
from cryptography.x509.oid import ExtensionOID, NameOID

from fluxvault.extensions import FluxVaultExtensions
from fluxvault.helpers import get_app_and_component_name
from fluxvault.log import log
from fluxvault.registrar import FluxAgentRegistrar, FluxPrimaryAgent, FluxSubAgent


class FluxAgentException(Exception):
    pass


class FluxAgent:
    """Runs on Flux nodes - waits for connection from FluxKeeper"""

    def __init__(
        self,
        bind_address: str = "0.0.0.0",
        bind_port: int = 8888,
        enable_registrar: bool = False,
        registrar: FluxAgentRegistrar | None = None,
        extensions: FluxVaultExtensions = FluxVaultExtensions(),
        working_dir: str = tempfile.gettempdir(),
        whitelisted_addresses: list = ["127.0.0.1"],
        verify_source_address: bool = False,
        signed_vault_connections: bool = False,
        zelid: str = "",
        subordinate: bool = False,
        primary_agent: FluxPrimaryAgent | None = None,
    ):
        # ToDo: look at using __async__ instead of run_until_complete
        self.enable_registrar = enable_registrar
        self.extensions = extensions
        self.loop = asyncio.get_event_loop()
        self.zelid = zelid
        self.working_dir = working_dir
        self.subordinate = subordinate
        self.registrar = registrar
        self.signed_vault_connections = signed_vault_connections
        self.bind_address = bind_address
        self.bind_port = bind_port
        self.whitelisted_addresses = whitelisted_addresses
        self.verify_source_address = verify_source_address
        self.primary_agent = primary_agent
        self.component_name, self.app_name = get_app_and_component_name()

        log.info(f"Component name: {self.component_name}, App name: {self.app_name}")

        if not self.signed_vault_connections and not self.verify_source_address:
            # Must verify source address as a minimum
            self.verify_source_address = True

        self.raise_on_state_errors()
        self.register_extensions()
        self.setup_registrar()
        self.loop.run_until_complete(self.setup_sub_agent())

        self.auth_provider = self.loop.run_until_complete(self.get_auth_provider())
        transport = EncryptedSocketServerTransport(
            bind_address,
            bind_port,
            whitelisted_addresses=whitelisted_addresses,
            verify_source_address=verify_source_address,
            auth_provider=self.auth_provider,
        )
        self.rpc_server = RPCServer(transport, JSONRPCProtocol(), self.extensions)

    @staticmethod
    async def get_app_owner_zelid(app_name: str) -> str:
        # ToDo: move this to helpers
        async with ClientSession() as session:
            async with session.get(
                f"https://api.runonflux.io/apps/appowner?appname={app_name}"
            ) as resp:
                data = await resp.json()
                zelid = data.get("data", "")
        return zelid

    async def setup_sub_agent(self):
        if self.subordinate:
            self.sub_agent = FluxSubAgent(
                self.component_name,
                self.app_name,
                self.primary_agent,
                address=self.bind_address,
            )
            await self.sub_agent.register_with_primary_agent()

    def setup_registrar(self):
        if self.enable_registrar and not self.registrar:
            self.registrar = FluxAgentRegistrar()

    def raise_on_state_errors(self):
        """Minimal tests to ensure we are good to run"""
        try:
            os.listdir(self.working_dir)
        except Exception as e:
            raise FluxAgentException(f"Error accessing working directory: {e}")

        if self.verify_source_address and not self.whitelisted_addresses:
            raise ValueError(
                "Whitelisted addresses must be provided if not signing connections"
            )

        if self.subordinate and not self.primary_agent:
            raise ValueError("Primary agent must be provided if subordinate")

    def register_extensions(self):
        self.extensions.add_method(self.get_all_object_hashes)
        self.extensions.add_method(self.write_objects)
        self.extensions.add_method(self.remove_objects)
        self.extensions.add_method(self.get_methods)
        self.extensions.add_method(self.get_subagents)
        self.extensions.add_method(self.generate_csr)
        self.extensions.add_method(self.install_cert)
        self.extensions.add_method(self.install_ca_cert)
        self.extensions.add_method(self.upgrade_to_ssl)
        self.extensions.add_method(self.load_plugins)
        self.extensions.add_method(self.list_server_details)
        self.extensions.add_method(self.connect_shell)
        self.extensions.add_method(self.disconnect_shell)
        self.extensions.add_method(self.get_state)
        self.extensions.add_method(self.get_directory_hashes)

        # self.extensions.add_method(self.run_entrypoint)
        # self.extensions.add_method(self.extract_tar)

    async def get_auth_provider(self):
        auth_provider = None
        if self.signed_vault_connections:
            # this is solely for testing without an app (outside of a Fluxnode)
            if self.zelid:
                address = self.zelid
            else:
                address = await self.get_app_owner_zelid(self.app_name)
            log.info(f"App zelid is: {address}")
            auth_provider = SignatureAuthProvider(address=address)
        return auth_provider

    def run(self):
        if self.enable_registrar:
            self.loop.create_task(self.registrar.start_app())
            log.info(
                f"Sub agent http server running on port {self.registrar.bind_port}"
            )

        task = self.loop.create_task(self.rpc_server.serve_forever())

        try:
            self.loop.run_forever()
        except Exception as e:
            # should this be finally: ?
            # this is when child fork closes (no loop)
            # OSError
            task.cancel()
        else:
            if self.enable_registrar:
                self.loop.run_until_complete(self.registrar.cleanup())

    async def run_async(self):
        if self.enable_registrar:
            self.loop.create_task(self.registrar.start_app())
            log.info(
                f"Sub agent http server running on port {self.registrar.bind_port}"
            )

        self.loop.create_task(self.rpc_server.serve_forever())

    async def upgrade_to_ssl(self):
        cert = tempfile.NamedTemporaryFile()
        key = tempfile.NamedTemporaryFile()
        ca_cert = tempfile.NamedTemporaryFile()
        with open(cert.name, "wb") as f:
            f.write(self.cert)
        with open(key.name, "wb") as f:
            f.write(self.key)
        with open(ca_cert.name, "wb") as f:
            f.write(self.ca_cert)

        log.info("Upgrading connection to SSL")
        context = ssl.SSLContext(ssl.PROTOCOL_TLS_SERVER)
        context.load_cert_chain(cert.name, keyfile=key.name)
        context.load_verify_locations(cafile=ca_cert.name)
        context.check_hostname = False
        context.verify_mode = ssl.VerifyMode.CERT_REQUIRED
        # context.set_ciphers("ECDHE-ECDSA-AES256-GCM-SHA384:ECDHE-RSA-AES256-GCM-SHA384")

        cert.close()
        key.close()
        ca_cert.close()

        transport = EncryptedSocketServerTransport(
            self.bind_address,
            self.bind_port + 1,
            whitelisted_addresses=self.whitelisted_addresses,
            verify_source_address=self.verify_source_address,
            auth_provider=self.auth_provider,
            ssl=context,
        )
        await self.rpc_server.transport.stop_server()
        log.info("Non SSL RPC server stopped")
        self.rpc_server = RPCServer(transport, JSONRPCProtocol(), self.extensions)
        self.loop.create_task(self.rpc_server.serve_forever())

    def cleanup(self):
        # ToDo: look at cleanup for rpc server too
        if self.registrar:
            self.loop.run_until_complete(self.registrar.cleanup())

    def opener(self, path, flags):

        return os.open(path, flags, 0o777)

    ### EXTERNAL METHODS CALLED FROM KEEPER BELOW HERE ###

    def get_methods(self) -> dict:
        """Returns methods available for the keeper to call"""
        return {k: v.__doc__ for k, v in self.extensions.method_map.items()}

    def get_state(self) -> dict:
        methods = self.get_methods()
        plugins = self.extensions.list_plugins()
        primary_agent = self.primary_agent.to_dict() if self.primary_agent else None
        return {
            "methods": methods,
            "plugins": plugins,
            "component_name": self.component_name,
            "app_name": self.app_name,
            "enable_registrar": self.enable_registrar,
            "zelid": self.zelid,
            "working_dir": self.working_dir,
            "subordinate": self.subordinate,
            "signed_vault_connections": self.signed_vault_connections,
            "bind_address": self.bind_address,
            "bind_port": self.bind_port,
            "whitelisted_addresses": self.whitelisted_addresses,
            "verify_source_address": self.verify_source_address,
            "primary_agent": primary_agent,
        }

    async def crc_file(self, filename: Path, crc: int) -> int:
        async with aiofiles.open(filename, "rb") as f:
            data = await f.read()
            crc = binascii.crc32(data, crc)

        return crc

    async def crc_directory(self, directory: Path, crc: int) -> int:
        crc = binascii.crc32(directory.name.encode(), crc)
        for path in sorted(directory.iterdir(), key=lambda p: str(p).lower()):
            crc = binascii.crc32(path.name.encode(), crc)

            if path.is_file():
                crc = await self.crc_file(path, crc)
            elif path.is_dir():
                crc = await self.crc_directory(path, crc)
        return crc

    async def get_object_crc(self, path: str) -> int:
        p = Path(path)

        if not p.is_absolute():
            p = self.working_dir / p

        if not p.exists():
            crc = 0

        elif p.is_dir():
            crc = await self.crc_directory(p, 0)

        elif p.is_file():
            crc = await self.crc_file(p, 0)

        return {"name": path, "crc32": crc}

    async def get_all_object_hashes(self, objects: list) -> list:
        """Returns the crc32 for each object that is being managed"""
        log.info(f"Returning crc's for {len(objects)} object(s)")
        tasks = []
        for obj in objects:
            tasks.append(self.loop.create_task(self.get_object_crc(obj)))
        results = await asyncio.gather(*tasks)
        return results

    async def get_file_hash(self, file: Path):
        crc = await self.crc_file(file, 0)
        return {str(file): crc}

    async def get_directory_hashes(self, dir: str):
        """Hashes up all files in a specific directory. if
        give relative path, out working dir is base path. Need
        to remove this again for each hash to give back common path format"""
        hashes = {}
        p = Path(dir)
        path_relative_to_workdir = False
        try:
            if not p.is_absolute():
                p = self.working_dir / p
                path_relative_to_workdir = True

            if not p.exists():
                return hashes

            crc = binascii.crc32(p.name.encode())

            hashes.update({str(p): crc})
            for path in sorted(p.iterdir(), key=lambda p: str(p).lower()):
                if path.is_dir():
                    hashes.update(await self.get_directory_hashes(str(path)))

                elif path.is_file():
                    hashes.update(await self.get_file_hash(path))

            if path_relative_to_workdir:
                hashes = {
                    str(Path(k).relative_to(self.working_dir)): v
                    for k, v in hashes.items()
                }
        except Exception as e:
            print(repr(e))
            raise

        return hashes

    async def remove_object(self, obj: str):
        p = Path(obj)

        if not p.is_absolute():
            p = self.working_dir / p

        if p.exists():
            if p.is_dir():
                await aioshutil.rmtree(p)
            elif p.is_file():
                await aiofiles_os.remove(p)

    async def write_object(self, obj: dict):
        # ToDo: brittle file path
        # ToDo: catch file PermissionError etc

        executable = False  # pass this in dict in future

        if isinstance(obj["data"], bytes):
            mode = "wb"
        elif isinstance(obj["data"], str):
            mode = "w"
        else:
            raise ValueError("Data written must be either str or bytes")

        p = Path(obj["path"])

        if not p.is_absolute():
            p = self.working_dir / p

        p.parent.mkdir(parents=True, exist_ok=True)

        # this will make the file being written executable
        opener = self.opener if executable else None

        if obj["is_dir"]:
            p.mkdir(parents=True, exist_ok=True)
            if not obj["data"]:
                return

        if obj.get("uncompressed", False):
            try:
                async with aiofiles.open(p, mode=mode, opener=opener) as file:
                    await file.write(obj["data"])
            # ToDo: tighten this up
            except Exception as e:
                log.error(repr(e))

        else:  # tarball
            fh = io.BytesIO(obj["data"])
            try:
                with tarfile.open(fileobj=fh, mode="r|bz2") as tar:
                    tar.extractall(str(p))
                return
            except Exception as e:
                print(f"Tarfile error: {repr(e)}")

    async def get_subagents(self):
        agents = {}
        if self.registrar:
            agents = {v.dns_name: v.as_dict() for v in self.registrar.sub_agents}
        return {"sub_agents": agents}

    async def write_objects(self, objects: list):
        """Will write to disk any file provided, in the format {"name": <content>}"""
        # ToDo: this should be tasks
        for obj in objects:
            log.info(f"Writing object {obj['path']}")
            await self.write_object(obj)

    async def remove_objects(self, objects: list):
        for obj in objects:
            log.info(f"Removing object {obj}")
            await self.remove_object(obj)

    async def connect_shell(self, peer):
        # peer is the source ip, host
        # json converts tuple to list
        peer = tuple(peer)

        child_pid, fd = pty.fork()
        if child_pid == 0:

            # Child process
            while log.hasHandlers():
                log.removeHandler(log.handlers[0])

            # trying to suppress log messages?!? Suspect it is this process
            # somehow sending data down the pipe to the remote SSL end causing
            # problems decoding ssl every once in a while
            try:
                subprocess.run("zsh")
            except:
                pass

        else:
            # Parent process

            # the peer may be the jumphost instead of actual browser,
            # we have no way of identifying if they are a jumphost so we
            # pass the RPCClient id of the originating request
            try:
                self.rpc_server.transport.attach_pty(child_pid, fd, peer)
                await self.rpc_server.transport.proxy_pty(peer)
            except Exception as e:
                print("In connect_shell")
                print(repr(e))

    async def disconnect_shell(self, peer):
        log.info(f"Disconnecting shell for peer: {peer[0]}")
        peer = tuple(peer)  # json convert to list
        self.rpc_server.transport.detach_pty(peer)

    def list_server_details(self):
        return {
            "working_dir": self.working_dir,
            "plugins": self.extensions.list_plugins(),
            "registrar_enabled": self.enable_registrar,
        }

    async def load_plugins(self, directory: str):
        p = Path(directory)

        if not p.is_absolute():
            p = self.working_dir / p

        log.info(f"loading plugins from directory {p}")

        # print(os.getcwd())
        sys.path.append(str(p))

        # or p.stat().st_size == 0:
        p.mkdir(parents=True, exist_ok=True)
        # if not p.exists():
        #     log.error("Plugin directory does not exist, skipping")
        #     return
        plugins = [
            f.rstrip(".py") for f in os.listdir(p) if os.path.isfile(os.path.join(p, f))
        ]

        log.debug(f"Plugins available: {plugins}")

        for f in plugins:
            importlib.invalidate_caches()
            # ToDo: wrap try / except
            try:
                plugin = importlib.import_module(f)
            except Exception as e:
                log.error(e)
            plugin = plugin.plugin
            if isinstance(plugin, FluxVaultExtensions):
                if plugin.required_packages:
                    try:
                        subprocess.run(
                            [sys.executable, "-m", "pip", "install"]
                            + plugin.required_packages,
                            check=True,
                            stdout=subprocess.DEVNULL,
                            stderr=subprocess.STDOUT,
                        )
                    except subprocess.CalledProcessError:
                        log.error("Error loading extensions packages, skipping")
                        return
                self.extensions.add_plugin(plugin)
                log.info(f"Plugin {plugin.plugin_name} loaded")
            else:
                log.error("Plugin load error... skipping")

    async def generate_csr(self):
        # ToDo: this needs to include a Fluxnode identifier
        altname = f"{self.component_name}.{self.app_name}.com"

        log.info(f"Generating CSR with altname {altname}")

        key = rsa.generate_private_key(
            public_exponent=65537,
            key_size=2048,
        )
        self.key = key.private_bytes(Encoding.PEM, PrivateFormat.PKCS8, NoEncryption())

        # public = key.public_key().public_bytes(
        #     Encoding.PEM, PublicFormat.SubjectPublicKeyInfo
        # )

        csr = (
            x509.CertificateSigningRequestBuilder()
            .subject_name(
                x509.Name(
                    [
                        x509.NameAttribute(NameOID.COMMON_NAME, altname),
                    ]
                )
            )
            .add_extension(
                x509.SubjectAlternativeName(
                    [
                        x509.DNSName(altname),
                    ]
                ),
                critical=False,
            )
            .sign(key, hashes.SHA256())
        )
        return {"csr": csr.public_bytes(Encoding.PEM)}

    async def install_cert(self, cert_bytes: bytes):
        self.cert = cert_bytes
        cert = x509.load_pem_x509_certificate(cert_bytes)
        issuer = cert.issuer.get_attributes_for_oid(NameOID.COMMON_NAME)[0].value
        san = cert.extensions.get_extension_for_oid(
            ExtensionOID.SUBJECT_ALTERNATIVE_NAME
        ).value.get_values_for_type(x509.DNSName)
        log.info(f"Installing cert from issuer {issuer} with Alt names {san}")

        # ToDo: this timing seems a bit off?
        await self.sub_agent.update_local_agent(enrolled=True)

    async def upgrade_connection(self):
        self.rpc_server.transport.upgrade_socket()

    async def install_ca_cert(self, cert_bytes: bytes):
        log.info("Installing CA cert")
        self.ca_cert = cert_bytes

    def extract_tar(self, file, target_dir):
        Path(target_dir).mkdir(parents=True, exist_ok=True)

        try:
            tar = tarfile.open(file)
            tar.extractall(target_dir)
            tar.close()
        # ToDo: Fix
        except Exception as e:
            log.error(repr(e))

    async def run_entrypoint(self, entrypoint: str):
        # ToDo: don't use shell
        proc = await asyncio.create_subprocess_shell(entrypoint)

        await proc.communicate()



# Standard library
import asyncio
import functools
import shutil
from dataclasses import dataclass, field
from functools import reduce
from pathlib import Path
from typing import Callable

# 3rd party
import aiohttp
import cryptography
from fluxrpc.auth import SignatureAuthProvider
from fluxrpc.client import RPCClient
from fluxrpc.exc import MethodNotFoundError
from fluxrpc.protocols.jsonrpc import JSONRPCProtocol
from fluxrpc.transports.socket.client import EncryptedSocketClientTransport
from cryptography.x509.oid import NameOID
from ownca import CertificateAuthority
from ownca.exceptions import OwnCAInvalidCertificate

# this package
from fluxvault.app_init import setup_filesystem_and_wallet
from fluxvault.extensions import FluxVaultExtensions
from fluxvault.fluxapp_config import FileSystemeGroup, FileSystemEntry, FluxAppConfig
from fluxvault.fluxkeeper_gui import FluxKeeperGui
from fluxvault.helpers import (
    FluxVaultKeyError,
    SyncStrategy,
    bytes_to_human,
    manage_transport,
    size_of_object,
    tar_object,
)
from fluxvault.log import log


@dataclass
class FluxVaultContext:
    agents: dict
    storage: dict = field(default_factory=dict)


class FluxKeeper:
    """Runs in your protected environment. Provides runtime
    data to your vulnerable services in a secure manner

    The end goal is to be able to secure an application's private data where visibility
    of that data is restricted to the application owner
    """

    # GUI hidden via cli, no where near ready, should probably disable
    def __init__(
        self,
        apps_config: list[FluxAppConfig],
        vault_dir: Path,
        gui: bool = False,
    ):
        self.apps_config = apps_config
        # ToDo: configurable port
        self.gui = FluxKeeperGui("127.0.0.1", 7777, self)

        self.loop = asyncio.get_event_loop()
        self.managed_apps: list[FluxAppManager] = []
        self.vault_dir = vault_dir
        self.root_dir = setup_filesystem_and_wallet(self.vault_dir)

        log.info(f"App Data directory: {self.root_dir}")
        log.info(f"Vault directory: {self.vault_dir}")

        self.init_certificate_authority()
        self.configure_apps()

        if gui:
            self.start_gui()

    def init_certificate_authority(self):
        common_name = "keeper.fluxvault.com"

        self.ca = CertificateAuthority(
            ca_storage=f"{str(self.root_dir / 'ca')}", common_name="Fluxvault Keeper CA"
        )

        try:
            cert = self.ca.load_certificate(common_name)
        except OwnCAInvalidCertificate:
            cert = self.ca.issue_certificate(common_name, dns_names=[common_name])

        self.cert = cert.cert_bytes
        self.key = cert.key_bytes
        self.ca_cert = self.ca.cert_bytes

    def configure_apps(self):
        for app_config in self.apps_config:
            app_config.update_root_dir(self.vault_dir / app_config.name)
            flux_app = FluxAppManager(self, app_config)
            self.managed_apps.append(flux_app)

    def start_gui(self):
        self.loop.run_until_complete(self.gui.start())

    async def manage_apps(self):
        for app in self.managed_apps:
            while True:
                await app.run_agent_tasks()
                if app.app_config.run_once:
                    break
                log.info(
                    f"sleeping {app.app_config.polling_interval} seconds for app {app.app_config.name}..."
                )
                await asyncio.sleep(app.app_config.polling_interval)


class FluxAppManager:
    def __init__(
        self,
        keeper: FluxKeeper,
        config: FluxAppConfig,
        extensions: FluxVaultExtensions = FluxVaultExtensions(),
    ):
        self.keeper = keeper
        self.app_config = config
        self.agents = []
        self.extensions = extensions
        self.network_state = {}

        self.build_agents()
        self.register_extensions()

    def __iter__(self):
        yield from self.agents

    def __len__(self):
        return len(self.agents)

    @staticmethod
    async def get_agent_ips(app_name: str) -> list:
        url = f"https://api.runonflux.io/apps/location/{app_name}"
        timeout = aiohttp.ClientTimeout(connect=10)
        retries = 3

        # look at making session appwide
        async with aiohttp.ClientSession() as session:
            for n in range(retries):
                try:
                    async with session.get(url, timeout=timeout) as resp:
                        if resp.status in [429, 500, 502, 503, 504]:
                            log.error(f"bad response {resp.status}... retrying")
                            continue

                        data = await resp.json()
                        break

                except aiohttp.ClientConnectorError:
                    log.error(f"Unable to connect to {url}... retrying")
                    await asyncio.sleep(n)
                    continue

        node_ips = []
        if data.get("status") == "success":
            nodes = data.get("data")
            for node in nodes:
                ip = node["ip"].split(":")[0]
                node_ips.append(ip)
        else:
            log.error("Return status from Flux api was not successful for agent IPs")

        return node_ips

    def add(self, agent: RPCClient):
        self.agents.append(agent)

    def get_by_id(self, id):
        for agent in self.agents:
            if agent.id == id:
                return agent

    def get_by_socket(self, socket: tuple):
        for agent in self.agents:
            if not agent.connected:
                continue

            local = agent.transport.writer.get_extra_info("sockname")
            if local == socket:
                return agent

    def proxied_agents(self):
        # return filter(lambda x: x.is_proxied, self.agents)
        for agent in self.agents:
            if agent.is_proxied:
                yield agent

    def agent_ids(self) -> list:
        return [x.id for x in self.agents]

    def primary_agents(self) -> filter:
        # return [x for x in self.agents if not x.is_proxied]
        return list(filter(lambda x: not x.is_proxied, self.agents))

    def build_agents(self):
        agent_ips = (
            self.app_config.agent_ips
            if self.app_config.agent_ips
            else self.get_agent_ips(self.app_config.name)
        )

        if not self.app_config.signing_key and self.app_config.sign_connections:
            raise ValueError("Signing key must be provided if signing connections")

        auth_provider = None
        if self.app_config.sign_connections and self.app_config.signing_key:
            auth_provider = SignatureAuthProvider(key=self.app_config.signing_key)

        for ip in agent_ips:
            transport = EncryptedSocketClientTransport(
                ip,
                self.app_config.comms_port,
                auth_provider=auth_provider,
                proxy_target="",
                on_pty_data_callback=self.keeper.gui.pty_output,
                on_pty_closed_callback=self.keeper.gui.pty_closed,
            )
            flux_agent = RPCClient(
                JSONRPCProtocol(), transport, (self.app_config.name, ip, "fluxagent")
            )
            self.add(flux_agent)

    def register_extensions(self):
        self.extensions.add_method(self.get_all_agents_methods)
        self.extensions.add_method(self.poll_all_agents)

    def get_methods(self):
        """Returns methods available for the keeper to call"""
        return {k: v.__doc__ for k, v in self.extensions.method_map.items()}

    def get_all_agents_methods(self) -> dict:
        return self.keeper.loop.run_until_complete(self._get_agents_methods())

    @manage_transport
    async def get_agent_method(self, agent: RPCClient):
        agent_proxy = agent.get_proxy()
        methods = await agent_proxy.get_methods()

        return {agent.id: methods}

    async def _get_agents_methods(self) -> dict:
        """Queries every agent and returns a list describing what methods can be run on
        each agent"""
        tasks = []
        for agent in self.agents:
            task = asyncio.create_task(self.get_agent_method(agent))
            tasks.append(task)

        results = await asyncio.gather(*tasks)
        return reduce(lambda a, b: {**a, **b}, results)

    @manage_transport
    async def get_state(self, agent: RPCClient):
        proxy = agent.get_proxy()
        self.network_state.update({agent.id: await proxy.get_state()})

    @staticmethod
    def get_extra_objects(
        managed_object: FileSystemEntry,
        local_hashes: dict[str, int],
        remote_hashes: dict[str, int],
    ) -> tuple[list[Path], int]:
        count = 0
        extras = []

        for remote_name in remote_hashes:
            remote_name = Path(remote_name)

            with_remote_prefix = str(
                remote_name.relative_to(managed_object.remote_path)
            )

            local_name = local_hashes.get(with_remote_prefix, None)

            if not local_name:
                count += 1
                if not extras:
                    extras.append(remote_name)

                extras = FileSystemeGroup.filter_hierarchy(remote_name, extras)

        return (extras, count)

    @staticmethod
    def get_missing_or_modified_objects(
        managed_object: FileSystemEntry,
        local_hashes: dict[str, int],
        remote_hashes: dict[str, int],
    ) -> tuple[list[Path], int, int]:
        # can't use zip here as we don't know remote lengths
        # set would work for filenames but not hashes
        # iterate hashes and find missing / modified objects

        missing = 0
        modified = 0
        candidates: list[Path] = []

        for name, local_crc in local_hashes.items():
            # these local hashes have been formatted in "common" format
            name = Path(name)

            remote_crc = remote_hashes.get(str(managed_object.remote_path / name), None)
            if not remote_crc:
                missing += 1

            elif remote_crc != local_crc:
                modified += 1

            if not remote_crc or remote_crc != local_crc:
                # if we are the only candidate, our name is relative to ourself so we just continue
                if not candidates:
                    candidates.append(name)

                # it will work without returning candidates here, but more clear this way
                candidates = FileSystemeGroup.filter_hierarchy(name, candidates)

        return (candidates, missing, modified)

    @manage_transport
    async def sync_objects(self, agent):
        log.debug(f"Contacting Agent {agent.id} to check if files required")
        # component_config = self.app_config.merge_global_into_component(agent.id[2])

        component_config = self.app_config.get_component(agent.id[2])
        remote_paths = []
        if not component_config:
            log.warn(
                f"No config found for component {agent.id[2]}, this component will only get globally specified files"
            )
        else:
            # await component_config.file_manager.flatten_dirs()
            remote_paths.extend(component_config.file_manager.expanded_remote_paths())

        # await self.app_config.file_manager.flatten_dirs()
        # remote_paths.extend(self.app_config.file_manager.absolute_remote_paths())

        agent_proxy = agent.get_proxy()
        # these get returned in whatever format we gave in remote_paths
        objects = await agent_proxy.get_all_object_hashes(remote_paths)

        log.debug(f"Agent {agent.id} remote file CRCs: {objects}")

        if not objects:
            log.warn(f"No objects to sync for {agent.id} specified... skipping!")
            return

        objects_to_write = []
        for obj in objects:
            remote_path = Path(obj["name"])
            # this is broken. They may not have had a component config and just using global files
            # or maybe just make it so you have to specify a component
            managed_object = (
                component_config.file_manager.get_object_by_remote_path(remote_path)
                if component_config
                else None
            )
            # print("managed object!!!!", managed_object)

            if not managed_object:
                log.warn(f"managed object: {remote_path} not found in component config")
                return

            managed_object.remote_crc = obj["crc32"]
            managed_object.compare_objects()

            if (
                managed_object.sync_strategy == SyncStrategy.ALLOW_ADDS
                and managed_object.is_dir()
                and not managed_object.in_sync
                and managed_object.validated_remote_crc == managed_object.remote_crc
            ):
                continue

            if (
                managed_object.sync_strategy == SyncStrategy.ENSURE_CREATED
                and managed_object.is_dir()
                and managed_object.remote_object_exists
            ):
                continue

            if not managed_object.in_sync and managed_object.is_dir():
                remote_path = managed_object.expanded_remote_path()
                # these are in our format (I think) so managed_object.remote_path / x
                remote_hashes = await agent_proxy.get_directory_hashes(remote_path)
                local_hashes = managed_object.get_directory_hashes()

                candidates, missing, modified = self.get_missing_or_modified_objects(
                    managed_object, local_hashes, remote_hashes
                )

                extra_objects, unknown = self.get_extra_objects(
                    managed_object, local_hashes, remote_hashes
                )

                log.info(
                    f"{missing} missing object(s), {modified} modified object(s) and {unknown} extra object(s)... fixing"
                )

                if (
                    managed_object.sync_strategy == SyncStrategy.STRICT
                    and extra_objects
                ):
                    # we need to remove extra objects
                    to_delete = [str(x) for x in extra_objects]
                    await agent_proxy.remove_objects(to_delete)
                    log.info(
                        f"Sync strategy set to {SyncStrategy.STRICT.name}, deleting extra objects..."
                    )
                elif SyncStrategy.ALLOW_ADDS:
                    managed_object.validated_remote_crc = managed_object.remote_crc

                log.info(
                    f"Deltas resolved... {len(candidates)} object(s) need to be resynced"
                )

                # candidate is in common form
                for candidate in candidates:
                    abs_local_path = managed_object.local_workdir / candidate

                    size = size_of_object(abs_local_path)

                    # ToDo: configurable
                    ONE_MB = 1048576 * 1
                    uncompressed = False
                    if abs_local_path.is_file():
                        if size > ONE_MB:
                            log.info(
                                f"File {candidate} with size {bytes_to_human(size)} is larger than uncompressed limit... compressing"
                            )
                            data = tar_object(abs_local_path)
                        else:  # < 1MB
                            uncompressed = True
                            with open(abs_local_path, "rb") as f:
                                data = f.read()

                    elif abs_local_path.is_dir():
                        data = tar_object(abs_local_path)

                    path_str = str(managed_object.remote_path / candidate)

                    objects_to_write.append(
                        {
                            "path": path_str,
                            "data": data,
                            "is_dir": abs_local_path.is_dir(),
                            "uncompressed": uncompressed,
                        }
                    )
                managed_object.in_sync = True
                managed_object.remote_object_exists = True
                managed_object.created = True

        # this is now files only
        objects_to_write.extend(component_config.file_manager.objects_to_agent_list())

        # print("o2w", objects_to_write)
        if objects_to_write:
            agent_proxy.notify()
            # ToDo: this should return status
            await agent_proxy.write_objects(objects=objects_to_write)
            # agent_proxy.one_way = False

    def poll_all_agents(self):
        self.keeper.loop.run_until_complete(self.run_agent_tasks())

    @manage_transport
    async def enroll_agent(self, agent: RPCClient):
        log.info(f"Enrolling agent {agent.id}")
        proxy = agent.get_proxy()
        res = await proxy.generate_csr()
        csr_bytes = res.get("csr")

        csr = cryptography.x509.load_pem_x509_csr(csr_bytes)
        hostname = csr.subject.get_attributes_for_oid(NameOID.COMMON_NAME)[0].value

        try:
            cert = self.keeper.ca.load_certificate(hostname)
            self.keeper.ca.revoke_certificate(hostname)
        except OwnCAInvalidCertificate:
            pass
        finally:
            # ToDo: there has to be a better way (don't delete cert)
            # start using CRL? Do all nodes need CRL - probably
            shutil.rmtree(f"ca/certs/{hostname}", ignore_errors=True)
            cert = self.keeper.ca.sign_csr(csr, csr.public_key())

        # This triggers agent to update registrar, it should probably
        # be it's own action
        await proxy.install_cert(cert.cert_bytes)
        await proxy.install_ca_cert(self.keeper.ca.cert_bytes)

        # proxy.one_way = True
        proxy.notify()
        await proxy.upgrade_to_ssl()
        # proxy.one_way = False

        # ToDo: function (don't mutate child properties)
        agent.transport.proxy_ssl = True
        agent.transport.proxy_port += 1

    @manage_transport
    async def enroll_subordinates(self, agent: RPCClient):
        agent_proxy = agent.get_proxy()
        resp = await agent_proxy.get_subagents()

        sub_names = [k for k in resp["sub_agents"]]
        log.info(f"Agent {agent.id} has the following subordinates: {sub_names}")
        address = agent.transport.address

        for target, payload in resp.get("sub_agents").items():
            role = payload.get("role")  # not implemented yet
            enrolled = payload.get("enrolled")

            # ToDo: check if already enrolled, may have rebooted

            if not enrolled:
                flux_agent = self.create_agent(address, target)
                await self.enroll_agent(flux_agent, True, True)
                self.add(flux_agent)

    def create_agent(
        self,
        address: str,
        proxy_target: str | None = None,
        auth_provider: SignatureAuthProvider | None = None,
    ) -> RPCClient:
        transport = EncryptedSocketClientTransport(
            address,
            self.app_config.comms_port,
            auth_provider=auth_provider,
            proxy_target=proxy_target,
            proxy_port=self.app_config.comms_port,
            proxy_ssl=False,
            cert=self.keeper.cert,
            key=self.keeper.key,
            ca=self.keeper.ca_cert,
            on_pty_data_callback=self.keeper.gui.pty_output,
            on_pty_closed_callback=self.keeper.gui.pty_closed,
        )
        flux_agent = RPCClient(
            JSONRPCProtocol(), transport, (self.app_config.name, address, proxy_target)
        )

        return flux_agent

    async def run_agent_tasks(self, tasks: list[Callable] = []):
        if not self.agents:
            log.info("No agents found... nothing to do")
            return

        # headless mode
        # ToDo: add cli `tasks` thingee
        if not tasks:
            tasks = [
                self.enroll_subordinates,
                self.sync_objects,
                self.get_state,
            ]

        for index, func in enumerate(tasks):
            log.info(f"Running task: {func.__name__}")
            # ToDo: if iscoroutine
            coroutines = []
            length = len(tasks)
            for agent in self.agents:
                connect = False
                disconnect = False
                if index == 0:
                    connect = True
                if index + 1 == length:
                    disconnect = True
                t = asyncio.create_task(func(agent, connect, disconnect))
                coroutines.append(t)
            try:
                await asyncio.gather(*coroutines)
            except FluxVaultKeyError as e:
                log.error(f"Exception from gather tasks: {repr(e)}")
                if self.keeper.gui:
                    await self.keeper.gui.set_toast(repr(e))

        if self.keeper.gui:
            await self.keeper.gui.app_state_update(
                self.app_config.name, self.network_state
            )

    def __getattr__(self, name: str) -> Callable:
        try:
            func = self.extensions.get_method(name)
        except MethodNotFoundError as e:
            raise AttributeError(f"Method does not exist: {e}")

        if func.pass_context:
            context = FluxVaultContext(self.agents)
            func = functools.partial(func, context)

        return func
